#include "raytracing.h"

Primitives Node::prim;

float clamp01(float v) {
  if (v < 0.0f) return 0.0f;
  if (v > 1.0f) return 1.0f;
  return v;
}

Vec3 Color(const Vec3& c) {
  return Vec3(clamp01(c.x), clamp01(c.y), clamp01(c.z));
}

Node::Node(const Mat4& m, const Vec3& c, float sp, float tr)
    : transfo(m), inv_transfo(glm::inverse(m)), Col(c), spec(sp), transp(tr) {}

Vec3 transfo_point(Vec3 p, Mat4 trans) {
  Vec4 new_p = trans * Vec4(p, 1.0f);
  return Vec3(new_p / new_p.w);
}

Vec3 transfo_dir(Vec3 d, Mat4 trans) {
  Vec4 new_d = trans * Vec4(d, 0.0f);
  return Vec3(new_d);
}

void Cube::draw_gl() const {
  Node::prim.draw_cube(this->transfo, this->Col);
  // DC
}
void Sphere::draw_gl() const {
  Node::prim.draw_sphere(this->transfo, this->Col);
  // DC
}
void Cylinder::draw_gl() const {
  Node::prim.draw_cylinder(this->transfo, this->Col);
  // DC
}

bool Cube::intersecte(const Vec3& O, const Vec3& D, Inter* I) {
  float min = -1.0f;
  float max = 1.0f;  // dimensions min et max en x, y, z du cube de longueur 2 à
                     // l'origine

  Vec3 O_inv = transfo_point(O, this->inv_transfo);
  Vec3 D_inv = transfo_dir(D, this->inv_transfo);

  float tmin = (min - O_inv.x) / D_inv.x;
  float tmax = (max - O_inv.x) / D_inv.x;

  if (tmin > tmax) std::swap(tmin, tmax);

  float tymin = (min - O_inv.y) / D_inv.y;
  float tymax = (max - O_inv.y) / D_inv.y;

  if (tymin > tymax) std::swap(tymin, tymax);

  if ((tmin > tymax) || (tymin > tmax)) return false;

  if (tymin > tmin) tmin = tymin;

  if (tymax < tmax) tmax = tymax;

  float tzmin = (min - O_inv.z) / D_inv.z;
  float tzmax = (max - O_inv.z) / D_inv.z;

  if (tzmin > tzmax) std::swap(tzmin, tzmax);

  if ((tmin > tzmax) || (tzmin > tmax)) return false;

  Vec3 P_inter;

  if (tzmin > tmin) tmin = tzmin;

  if (tzmax < tmax) tmax = tzmax;

  // choix du point d'intersection le plus proche
  if (tmax - tmin > 0.0f) {
    P_inter = O + tmin * D;
  } else {
    P_inter = O + tmax * D;
  }

  I->a_min = glm::distance(O, P_inter);
  I->node = this;
  I->O = P_inter;

  return true;
}

Vec3 Cube::normal(const Inter& I, const Vec3& P) {
  // AC. Pensez à normaliser la normale une fois le vecteur calculé.
  Vec3 t = transfo_point(I.O, this->inv_transfo);
  if (abs(t.x) >= 1. - 0.001f)
    return glm::normalize(transfo_dir(Vec3(t.x, 0, 0), this->transfo));
  if (abs(t.y) >= 1. - 0.001f)
    return glm::normalize(transfo_dir(Vec3(0, t.y, 0), this->transfo));
  if (abs(t.z) >= 1. - 0.001f)
    return glm::normalize(transfo_dir(Vec3(0, 0, t.z), this->transfo));
  return glm::normalize(t);
}

bool Sphere::intersecte(const Vec3& O, const Vec3& D, Inter* I) {
  Vec3 O_inv = transfo_point(O, this->inv_transfo);
  Vec3 D_inv = transfo_dir(D, this->inv_transfo);

  float a =
      (D_inv.x) * (D_inv.x) + (D_inv.y) * (D_inv.y) + (D_inv.z) * (D_inv.z);
  float b = 2.0f * D_inv.x * O_inv.x + 2.0f * D_inv.y * O_inv.y +
            2.0f * D_inv.z * O_inv.z;
  float c = (O_inv.x) * (O_inv.x) + (O_inv.y) * (O_inv.y) +
            (O_inv.z) * (O_inv.z) - 1.0f;

  float deter = b * b - 4 * a * c;

  if (deter > 0.0f) {
    // point d'intersection
    Vec3 P_inter;

    float t0 = (-b + sqrtf(deter)) / (2 * a);
    float t1 = (-b - sqrtf(deter)) / (2 * a);
    t0 < 0 ? t0 = -t0 : t0;
    t1 < 0 ? t1 = -t1 : t1;

    // choix du point d'intersection le plus proche
    if (t1 - t0 > 0.0f) {
      P_inter = O + t0 * D;
    } else {
      P_inter = O + t1 * D;
    }

    I->a_min = glm::distance(O, P_inter);
    I->node = this;
    I->O = P_inter;

    return true;
  } else if (deter == 0.0f) {
    float t = -b / (2 * a);
    Vec3 P_inter = O + t * D;

    I->a_min = glm::distance(O, P_inter);
    I->node = this;
    I->O = P_inter;

    return true;
  } else
    return false;
}

Vec3 Sphere::normal(const Inter& I, const Vec3& P) {
  Vec3 tr = transfo_point(I.O, this->inv_transfo);
  return glm::normalize(transfo_dir(tr, glm::inverseTranspose(this->transfo)));
}

bool Cylinder::intersecte(const Vec3& O, const Vec3& D, Inter* I) {
  Vec3 O_inv = transfo_point(O, this->inv_transfo);
  Vec3 D_inv = transfo_dir(D, this->inv_transfo);

  float a = D_inv.x * D_inv.x + D_inv.y * D_inv.y;
  float b = 2.0f * D_inv.x * O_inv.x + 2.0f * D_inv.y * O_inv.y;
  float c = (O_inv.x) * (O_inv.x) + (O_inv.y) * (O_inv.y) - 1.0f;

  float tr, tmp, deter = b * b - 4 * a * c;
  if (deter > 0.0f) {
    // point d'intersection
    Vec3 P_inter;

    float t0 = (-b - sqrtf(deter)) / (2 * a);
    float t1 = (-b + sqrtf(deter)) / (2 * a);
    if(t0 > t1) {
      tmp = t0;
      t0 = t1;
      t1 = tmp;
    }

    float z0 = O_inv.z + t0 * D_inv.z;
    float z1 = O_inv.z + t1 * D_inv.z;


    // choix du point d'intersection le plus proche
    if (t1 - t0 > 0.0f) {
      P_inter = O + t0 * D;
    } else {
      P_inter = O + t1 * D;
    }

    I->a_min = glm::distance(O, P_inter);
    I->node = this;
    I->O = P_inter;

    return true;
  } else if (deter == 0.0f) {
    float t = -b / (2 * a);
    Vec3 P_inter = O + t * D;

    I->a_min = glm::distance(O, P_inter);
    I->node = this;
    I->O = P_inter;

    return true;
  } else
    return false;
}

Vec3 Cylinder::normal(const Inter& I, const Vec3& P) {
  return BLANC;
}

template <typename F>
void BVH::intersecte(const Vec3& O, const Vec3& D, Inter* I, F func) {
  unsigned long i;  // , j;
  Inter I2;

  for (i = 0; i < nodes.size(); i++) {
    if (this->nodes[i]->intersecte(O, D, &I2)) {
      if (I2.a_min < I->a_min) {
        *I = I2;
      }
    }
  }
  /*BVH *bvh_sec = this;
  while(bvh_sec->children.size() > 0){
      for(i = 0; i < bvh_sec->children.size(); i++){
          for(j = 0; j < bvh_sec->children[i]->nodes.size(); j++){
              if(bvh_sec->children[i]->nodes[j].intersecte(O, D, I2)){
                  if(I2->a_min < I->a_min){
                      I = I2;
                  }
              }
          }
      }
      bvh_sec = bvh_sec->children[0];
  }*/
  // AC. Vous devez normalement changer le contenu de I.
}

const float Node::Epsilon = 0.0001f;

RTracer::RTracer(SimpleViewer* s) : sv(s), depth(0) {}

void RTracer::add_cube_bvh(BVH* b, const Mat4& m, const Vec3& color, float spec,
                           float tr) {
  b->nodes.push_back(new Cube(m, color, spec, tr));
  // DC
}
void RTracer::add_sphere_bvh(BVH* b, const Mat4& m, const Vec3& color,
                             float spec, float tr) {
  b->nodes.push_back(new Sphere(m, color, spec, tr));
  // DC
}
void RTracer::add_cylinder_bvh(BVH* b, const Mat4& m, const Vec3& color,
                               float spec, float tr) {
  b->nodes.push_back(new Cylinder(m, color, spec, tr));
  // DC
}

Vec3 RTracer::ColorRayBVH(const Vec3& P, const Vec3& V, int depth) {
  Inter I;
  I.a_min = std::numeric_limits<float>::max();
  I.node = nullptr;

  bvh->intersecte(P, V, &I, [](Node*) { return true; });

  if (I.node) {
    Vec3 norm = I.node->normal(I, Vec3(0.));
    Vec3 light = glm::normalize(this->posLum - I.O);
    float p = glm::dot(light, norm);
    p = clamp01(p);
    Vec3 f = I.node->Col * p;
    return Color(f);
  }
  return NOIR;

  // AC en lisant le résultat du test d'intersection dans I
}

void RTracer::calc_thr(QImage* im, int t) {
  for (int y = t; y < this->sv->height(); y += this->nbthr) {
    for (int x = 0; x < this->sv->width(); ++x) {
      qglviewer::Vec Pq = this->sv->camera()->unprojectedCoordinatesOf(
          qglviewer::Vec(x, y, 0.0));
      qglviewer::Vec Qq = this->sv->camera()->unprojectedCoordinatesOf(
          qglviewer::Vec(x, y, 1.0));
      Vec3 P(Pq[0], Pq[1], Pq[2]);
      Vec3 D(Qq[0] - Pq[0], Qq[1] - Pq[1], Qq[2] - Pq[2]);
      Vec3 C = this->ColorRayBVH(P, D, this->depth);
      // ^ ou ColorRay(...) : lance un rayon d'origine P et de direction D,
      // rebondissant this->depth fois.
      im->setPixel(
          x, y,
          QColor(int(C.r * 255.0f), int(C.g * 255.0f), int(C.b * 255.0f))
              .rgb());
    }
  }
  // DC
}

QLabel* RTracer::CalcImage(int depth) {
  QImage im(this->sv->width(), this->sv->height(), QImage::Format_RGB32);
  Vec4 SHADER_POSLUM(300, 1000, 1000, 1);
  this->posLum =
      Vec3(glm::inverse(this->sv->getCurrentModelViewMatrix()) * SHADER_POSLUM);
  this->depth = depth;
  std::vector<std::thread*> threads;
  threads.reserve(this->nbthr);

  for (int i = 0; i < this->nbthr; ++i) {
    // En C++ on utilise les threads comme ceci.
    // La fonction lambda passée en paramètre du constructeur est exécutée par
    // un thread.
    threads.push_back(
        new std::thread([i, &im, this]() { this->calc_thr(&im, i); }));
  }

  for (auto* t : threads) {
    // join fonctionne comme une barrière, où le thread principal attend que
    // chaque thread ait terminé pour le libérer. il est également possible
    // d'utiliser des mutex avec std::mutex, mais c'est une autre histoire.
    t->join();
    delete t;
  }

  QLabel* label_img = new QLabel(this->sv);
  label_img->setWindowFlags(Qt::Window);
  label_img->setPixmap(QPixmap::fromImage(im));
  return label_img;
  // DC
}

void draw_prim_bvh(BVH* b) {
  for (const auto* n : b->nodes) n->draw_gl();
  for (auto* c : b->children) draw_prim_bvh(c);
  // DC
}
